
auction <- function(agents, marketOrderCash=0, digits=2,
                    plotVols=FALSE, debug=FALSE)
{
  if(inherits(agents, "marketAgentList")) {
    sellsorig <- unlist(lapply(agents, `[[`, "sellAt"))
    buysorig <- unlist(lapply(agents, `[[`, "buyAt"))
    cashsorig <- unlist(lapply(agents, `[[`, "cash"))
    sharesorig <- unlist(lapply(agents, `[[`, "shares"))
  } else {
    sellsorig <- agents$sellAt
    buysorig <- agents$buyAt
    cashsorig <- agents$cash
    sharesorig <- agents$shares
  }
  buysorig <- buysorig[cashsorig > 0]
  cashsorig <- cashsorig[cashsorig > 0]
  sellsorig <- sellsorig[sharesorig > 0]
  sharesorig <- sharesorig[sharesorig > 0]

  defaultAnswer <- c(price=NA, volume=0, unmatched=0)

  if(!length(buysorig) || !length(sellsorig)) return(defaultAnswer)
  tint <- c(min(sellsorig), max(buysorig))
  if(tint[1L] > tint[2L]) return(defaultAnswer)

  shares <- sharesorig[sellsorig <= tint[2L]]
  sells <- sellsorig[sellsorig <= tint[2L]]
  cashs <- cashsorig[buysorig >= tint[1L]]
  buys <- buysorig[buysorig >= tint[1L]]

  dmult <- 10^digits
  ticks <- (ceiling(tint[1L] * dmult):floor(tint[2L] * dmult)) / dmult

  mbcash <- mscash <- 0
  if(marketOrderCash < 0) {
    mscash <- -marketOrderCash
  } else if(marketOrderCash > 0) {
    mbcash <- marketOrderCash
  }

  nticks <- length(ticks)
  tvolbuy <- tvolsell <- numeric(nticks)

  if(mbcash > 0) {
    for(i in nticks) {
      tt <- ticks[i]
      tbe <- buys >= tt
      tse <- sells <= tt
      tvolbuy[i] <- sum(floor(c(cashs[tbe], mbcash) / tt))
      tvolsell[i] <- sum(shares[tse]) + floor(mscash / tt)
    }
    if(tvolsell[nticks] * ticks[nticks] < mbcash) {
      newtint <- c(min(ticks), max(sellsorig))
      ticks <- (ceiling(newtint[1L] * dmult):floor(newtint[2L] * dmult)) /
        dmult
      sells <- sellsorig
      shares <- sharesorig
      nticks <- length(ticks)
      tvolbuy <- tvolsell <- numeric(nticks)
    }
  } else if(mscash > 0) {
    for(i in 1L) {
      tt <- ticks[i]
      tbe <- buys >= tt
      tse <- sells <= tt
      tvolbuy[i] <- sum(floor(c(cashs[tbe], mbcash) / tt))
      tvolsell[i] <- sum(shares[tse]) + floor(mscash / tt)
    }
    if(tvolbuy[1L] * ticks[1L] < mscash) {
      newtint <- c(min(buysorig), max(ticks))
      ticks <- (ceiling(newtint[1L] * dmult):floor(newtint[2L] * dmult)) /
        dmult
      ticks <- ticks[-1L]
      buys <- buysorig
      cashs <- cashsorig
      nticks <- length(ticks)
      tvolbuy <- tvolsell <- numeric(nticks)
    }
  }
  # actual loop
  for(i in seq_len(nticks)) {
    tt <- ticks[i]
    tbe <- buys >= tt
    tse <- sells <= tt
    tvolbuy[i] <- sum(floor(c(cashs[tbe], mbcash) / tt))
    tvolsell[i] <- sum(shares[tse]) + floor(mscash / tt)
  }


  if(plotVols) {
    matplot(ticks, cbind(tvolbuy, tvolsell), type="l", lwd=3, lty=1,
            xlab="Price", ylab="Volume")
  }
  volume <- pmin(tvolbuy, tvolsell)
  apoint <- which.max(volume)

  unmatched <- if(tvolsell[1L] > tvolbuy[1L]) {
    tvolbuy[1L] - tvolsell[1L]
  } else if(tvolsell[nticks] < tvolbuy[nticks]) {
    tvolbuy[nticks] - tvolsell[nticks]
  } else 0

  ans <- c(price=ticks[apoint], volume=max(volume[apoint], 0), unmatched=unmatched)
  anssup <- c(buyall=unname(ans["price"]), buypartprice=NA, buyfraction=NA,
              sellall=unname(ans["price"]), sellpartprice=NA, sellfraction=NA,
              marketcash=0, marketshares=0)
  if(mscash > 0) {
    anssup["marketshares"] <- -floor(mscash / ans["price"])
  } else if(mbcash > 0) {
    anssup["marketshares"] <- floor(mbcash / ans["price"])
  }
  anssup["marketcash"] <- -anssup["marketshares"] * ans["price"]

  if(ans["volume"] > 0) {
    vb <- tvolbuy[apoint]
    vs <- tvolsell[apoint]
    # browser()
    if(vs < vb) {
      # need to adjust buys
      apPlusOffset <- nticks - sum(vs >= tvolbuy[-1:-apoint])
      if(apPlusOffset < nticks) {
        anssup["buyall"] <- ticks[apPlusOffset + 1L]
        if(tvolbuy[apPlusOffset] > vs) {
          anssup["buypartprice"] <- ticks[apPlusOffset]
          anssup["buyfraction"] <- (vs - tvolbuy[apPlusOffset + 1L]) /
            (tvolbuy[apPlusOffset]- tvolbuy[apPlusOffset + 1L])
        }
      } else {
        anssup["buyall"] <- NA
        anssup["buypartprice"] <- ticks[nticks]
        anssup["buyfraction"] <- vs / tvolbuy[nticks]
      }
    } else if(vb < vs) {
      # need to adjust sells
      offset <- sum(vb <= tvolsell[1:apoint])
      if(apoint - offset > 0L) {
        anssup["sellall"] <- ticks[apoint - offset]
        if(tvolsell[apoint - offset + 1] > vb) {
          anssup["sellpartprice"] <- ticks[apoint - offset +  1L]
          anssup["sellfraction"] <- (vb - tvolsell[apoint - offset]) /
            (tvolsell[apoint - offset + 1L] - tvolsell[apoint - offset])
        }
      } else {
        anssup["sellall"] <- NA
        anssup["sellpartprice"] <- ticks[1L]
        anssup["sellfraction"] <- vb/tvolsell[1L]
      }
    }
  } else {
    ans["price"] <- NA
  }
  ans <- c(ans, anssup)
  if(any(is.infinite(ans))) {
    stop("infinite value in auction")
  }

  if(debug) recover()

  ans
}
